package animation;

import animation.Memory.MemoryType;
import animation.StackFrame.FrameType;

public class PseudoCodeProcessor {

    public static enum CodeStatus
    {
        UNFINISHED,
        BREAKPOINT,
        FINISHED
    }
    
    private String currentDebugOutput;
    private Memory mem;
    private PseudoCodeNode programPointer;
    
    public PseudoCodeProcessor( PseudoCodeNode tree )
    {
        mem = new Memory();
        mem.addFrame( new StackFrame( FrameType.FUNCTION ) );
        programPointer = tree;
        currentDebugOutput = "";
    }
    
    private CodeStatus selectNextNode( PseudoCodeNode next, PseudoCodeNode last )
    {
        programPointer = next;
        last.setSelected( false );
        switch( next.setSelected( true ) )
        {
        case CONTINUE:
            return CodeStatus.UNFINISHED;
        case SKIP:
            return forwardStepOver();
        case STOP:
            return CodeStatus.BREAKPOINT;
        }
        return CodeStatus.UNFINISHED;
    }
    
    public CodeStatus forwardStep()
    {
        if( programPointer == null )
            return CodeStatus.FINISHED;
        StackFrame before = mem.removeFrame();
        mem.addFrame( before );
        ControlFlow cf = programPointer.forwardStep( mem );
        currentDebugOutput = programPointer.getDebugOutput( mem );
        switch( cf.getStatus() )
        {
        case ControlFlow.STEP_INTO:
            if( mem.isDefined( "_returnTo" + programPointer.getId(), MemoryType.GLOBAL ) )
            {
                mem.declare( "_returnTo" + programPointer, mem.read( "_returnTo" + programPointer.getId(), MemoryType.GLOBAL ), MemoryType.LOCAL );
                mem.undeclare( "_returnTo" + programPointer.getId(), MemoryType.GLOBAL );
            }
            if( programPointer.children() == null )
                throw new IllegalStateException( "A Codeline without sublines tryd to make a STEP_INTO." );
            else
                return selectNextNode( (PseudoCodeNode)programPointer.getFirstChild(), programPointer );
        case ControlFlow.STEP_OVER:
            if( programPointer.getParent() == null )
                return CodeStatus.FINISHED;
            if( before.isDefined( "_returnTo" + programPointer.getId() ) )
            {
                PseudoCodeNode nextPC = before.<PseudoCodeNode>get( "_returnTo" + programPointer.getId() );
                before.undeclare( "_returnTo" + programPointer.getId() );
                return selectNextNode( nextPC, programPointer );
            }
            PseudoCodeNode nextPC = (PseudoCodeNode) ((PseudoCodeNode)programPointer.getParent()).getChildAfter( programPointer );
            if( nextPC == null )
                return selectNextNode( (PseudoCodeNode)programPointer.getParent(), programPointer );
            else
                return selectNextNode( nextPC, programPointer );
        case ControlFlow.CALL:
            mem.declare( "_returnTo" + cf.getFunction().getId(), programPointer, MemoryType.GLOBAL );
            return selectNextNode( cf.getFunction(), programPointer );
        }
        throw new IllegalStateException( "Unbekannte ControlFlow Aktion" );
    }
    
    public CodeStatus forwardStepOver()
    {
        if( programPointer == null )
            return CodeStatus.FINISHED;
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        do {
            status = forwardStep();
        } while( mem.getSize() > stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }
    
    public CodeStatus forwardStepOut()
    {
        if( programPointer == null )
            return CodeStatus.FINISHED;
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        do {
            status = forwardStep();
        } while( mem.getSize() >= stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }
    
    private CodeStatus selectBeforeNode( PseudoCodeNode next, PseudoCodeNode last )
    {
        programPointer = next;
        last.setSelected( false );
        switch( next.setSelected( true ) )
        {
        case CONTINUE:
            return CodeStatus.UNFINISHED;
        case SKIP:
            return backwardStepOver();
        case STOP:
            return CodeStatus.BREAKPOINT;
        }
        return CodeStatus.UNFINISHED;
    }
    
    public CodeStatus backwardStep()
    {
        if( programPointer == null )
            return CodeStatus.FINISHED;
        StackFrame before = mem.removeFrame();
        mem.addFrame( before );
        ControlFlow cf = programPointer.backwardStep( mem );
        currentDebugOutput = programPointer.getDebugOutput( mem );
        switch( cf.getStatus() )
        {
        case ControlFlow.STEP_INTO:
            if( mem.isDefined( "_returnTo" + programPointer.getId(), MemoryType.GLOBAL ) )
            {
                mem.declare( "_returnTo" + programPointer, mem.read( "_returnTo" + programPointer.getId(), MemoryType.GLOBAL ), MemoryType.LOCAL );
                mem.undeclare( "_returnTo" + programPointer.getId(), MemoryType.GLOBAL );
            }
            if( programPointer.children() == null )
                throw new IllegalStateException( "A Codeline without sublines tryd to make a STEP_INTO." );
            else
                return selectBeforeNode( (PseudoCodeNode)programPointer.getLastChild(), programPointer );
        case ControlFlow.STEP_OVER:
            if( programPointer.getParent() == null )
                return CodeStatus.FINISHED;
            if( before.isDefined( "_returnTo" + programPointer.getId() ) )
            {
                PseudoCodeNode nextPC = before.<PseudoCodeNode>get( "_returnTo" + programPointer.getId() );
                before.undeclare( "_returnTo" + programPointer.getId() );
                return selectBeforeNode( nextPC, programPointer );
            }
            PseudoCodeNode nextPC = (PseudoCodeNode) ((PseudoCodeNode)programPointer.getParent()).getChildBefore( programPointer );
            if( nextPC == null )
                return selectBeforeNode( (PseudoCodeNode) programPointer.getParent(), programPointer );
            else
                return selectBeforeNode( nextPC, programPointer );
        case ControlFlow.CALL:
            mem.declare( "_returnTo" + cf.getFunction().getId(), programPointer, MemoryType.GLOBAL );
            return selectBeforeNode( cf.getFunction(), programPointer );
        }
        throw new IllegalStateException( "Unbekannte ControlFlow Aktion" );
    }
    
    public CodeStatus backwardStepOver()
    {
        if( programPointer == null )
            return CodeStatus.FINISHED;
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        do {
            status = backwardStep();
        } while( mem.getSize() > stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }
    
    public CodeStatus backwardStepOut()
    {
        if( programPointer == null )
            return CodeStatus.FINISHED;
        int stackSize = mem.getSize();
        CodeStatus status = CodeStatus.UNFINISHED;
        do {
            status = backwardStep();
        } while( mem.getSize() >= stackSize && status == CodeStatus.UNFINISHED );
        return status;
    }
    
    public String getDebugOutput()
    {
        return currentDebugOutput;
    }
}
